local socket = require 'socket'
local try, protect, newtry  = socket.try, socket.protect, socket.newtry
local setmetatable = setmetatable
local string, table, type  = string, table, type
local insert = table.insert
local tonumber = tonumber

local print = print

module(...)

local MT -- metatable

local function cmd_respond(cln, cmd, ok_code)
    try(cln.conn:send(cmd .. '\r\n'))
    local resp = try(cln.conn:receive())
    local code = string.sub(resp, 1, 3)
    if string.match(ok_code, code) then
        return tonumber(code), resp
    else
        return nil, resp
    end
end


local function newclient(conf)
    local host, port, create
    if type(conf) == 'string' then 
        host = conf
        port = 119
        create = socket.tcp
    else
        host = conf.host
        port = conf.port or 119
        create = conf.create or socket.tcp
    end
    local cln = {}
    setmetatable(cln, MT)
    cln.conn = try(create())
    try(cln.conn:connect(host, port))
    cln.banner = try(cln.conn:receive())
    local code = string.sub(cln.banner, 1, 3)
    if code == "200" then
        cln.posting_ok = true
    elseif code == "201" then
        cln.posting_ok = false
    else
        cln.conn:close()
        return nil, cln.banner
    end
    return cln
end

local function get_data_table(cln, fnc)
    fnc = fnc or insert
    local data = {}
    while true do
        local line = try(cln.conn:receive())
        if line == "." then break end
        -- unstuff the line
        if string.sub(line, 1, 1) == "." then line = string.sub(line, 2) end
        fnc(data, line)
    end
    return data
end

local function get_data(cln)
    local data = get_data_table(cln)
    insert(data, "")
    return table.concat(data, "\r\n")
end

local function article(cln, msg_id)
    local cmd = "ARTICLE"
    if msg_id then
        cmd = "ARTICLE " .. msg_id
    end
    try(cmd_respond(cln, cmd, "220"))
    return get_data(cln)
end

local function body(cln, msg_id)
    local cmd = "BODY"
    if msg_id then
        cmd = "BODY " .. msg_id
    end
    try(cmd_respond(cln, cmd, "222"))
    return get_data(cln)
end

local function head(cln, msg_id)
    local cmd = "HEAD"
    if msg_id then
        cmd = "HEAD " .. msg_id
    end
    try(cmd_respond(cln, cmd, "221"))
    return get_data(cln)
end

local function stat(cln, msg_id)
    local cmd = "STAT"
    if msg_id then
        cmd = "STAT " .. msg_id
    end
    local code, resp =  try(cmd_respond(cln, cmd , "223"))
    local num, msg_id = string.match(resp, "%d+ (%d+) (<.+>)")
    return tonumber(num), msg_id
end

local function group(cln, grp)
    local cmd = "GROUP " .. grp
    local code, resp =  try(cmd_respond(cln, cmd , "211"))
    local count, first, last, grp = string.match(resp, "%d+ (%d+) (%d+) (%d+) (%S+)")
    return tonumber(count), tonumber(first), tonumber(last), grp
end

local function help(cln)
    local cmd = "HELP"
    try(cmd_respond(cln, cmd , "100"))
    return get_data(cln)
end

local function ihave(cln, msg_id, msg)
    local cmd = "IHAVE " .. msg_id
    try(cmd_respond(cln, cmd , "335"))
    try(cln.conn:send(msg))
    if string.sub(msg, -2) == "\r\n" then 
        cmd = "."
    else
        cmd = "\r\n."
    end
    return cmd_respond(cln, cmd, "235")
end

local function last(cln)
    local cmd = "LAST"
    local code, resp =  try(cmd_respond(cln, cmd , "223"))
    local msg_no, msg_id = string.match(resp, "%d+ (%d+) (<%S+>)")
    return tonumber(msg_no), msg_id
end

local function list(cln)
    local cmd = "LIST"
    try(cmd_respond(cln, cmd , "215"))
    return get_data_table(cln,
                          function (tab, line) 
                              local grp, max, min, post = string.match(line, "(%S+) (%d+) (%d+) (%S)")
                              if post == "y" then post=true else post=false end
                              tab[grp] = {tonumber(max), tonumber(min), post}
                          end)
end

local function next(cln)
    local cmd = "NEXT"
    local code, resp =  try(cmd_respond(cln, cmd , "223"))
    local msg_no, msg_id = string.match(resp, "%d+ (%d+) (<%S+>)")
    return tonumber(msg_no), msg_id
end

local function newgroups(cln, dd, tt, gmt, dist)
    local cmd = string.format("NEWGROUPS %s %s", dd, tt)
    if gmt then cmd = cmd .. " " .. gmt end
    if dist then cmd = cmd .. " " .. dist end
    try(cmd_respond(cln, cmd , "231"))
    return get_data_table(cln,
                          function (tab, line) 
                              local grp, max, min, post = string.match(line, "(%S+) (%d+) (%d+) (%S)")
                              if post == "y" then post=true else post=false end
                              tab[grp] = {tonumber(max), tonumber(min), post}
                          end)
end

local function newnews(cln, grp, dd, tt, gmt, dist)
    local cmd = string.format("NEWNEWS %s %s %s", grp, dd, tt)
    if gmt then cmd = cmd .. " " .. gmt end
    if dist then cmd = cmd .. " " .. dist end
    try(cmd_respond(cln, cmd , "230"))
    return get_data_table(cln)
end


local function post(cln, msg)
    local cmd = "POST"
    try(cmd_respond(cln, cmd , "340"))
    try(cln.conn:send(msg))
    if string.sub(msg, -2) == "\r\n" then 
        cmd = "."
    else
        cmd = "\r\n."
    end
    return cmd_respond(cln, cmd, "240")
end

local function quit(cln)
    local cmd = "QUIT"
    return cmd_respond(cln, cmd , "205")
end
    

local function slave(cln)
    local cmd = "SLAVE"
    return cmd_respond(cln, cmd , "202")
end

local function close(cln)
    return cln.conn:close()
end

new = protect(newclient)

MT = {
    article = protect(article),
    body = protect(body),
    head = protect(head),
    stat = protect(stat),
    group = protect(group),
    help = protect(help),
    ihave = protect(ihave),
    last = protect(last),
    list = protect(list),
    next = protect(next),
    newgroups = protect(newgroups),
    newnews = protect(newnews),
    post = protect(post),
    quit = protect(quit),
    slave = protect(slave),
    close = close,
}
MT.__index = MT
